"""
Plugin manager module.
Define the PluginManager class, the one that glues all the plugins and
the API together. Defines also a bunch of variables related to the
plugin env.
"""

import logging
import os
from typing import TYPE_CHECKING, Any, Callable, Type
from types import ModuleType
from importlib import import_module, machinery, util
from importlib.metadata import entry_points
from pathlib import Path
from os import path

from poezio import tabs, xdg
from poezio.core.structs import Command, Completion, GenericCompletion
from poezio.plugin import PluginAPI, BasePlugin
from poezio.config import config

if TYPE_CHECKING:
    from poezio.core.core import Core
    from poezio.windows.inputs import Input

log = logging.getLogger(__name__)


class PluginManager:
    """
    Plugin Manager
    Contains all the references to the plugins
    And keeps track of everything the plugin has done through the API.
    """

    rdeps: dict[str, set[str]] = {}
    modules: dict[str, ModuleType]
    plugins: dict[str, BasePlugin | None]
    commands: dict[str, dict[str, Command]]
    event_handlers: dict[str, list[tuple[str, Callable[..., Any]]]]
    tab_commands: dict[str, dict[str, list[tuple[
        str, Callable[[str], Any], str, GenericCompletion,
    ]]]]
    keys: dict[str, dict[str, Callable[..., Any]]]
    tab_keys: dict[str, dict[str, list[tuple[str, Callable[..., Any]]]]]

    def __init__(self, core: 'Core') -> None:
        self.core = core
        # module name -> module object
        self.modules = {}
        # module name -> plugin object
        self.plugins = {}
        # module name -> dict of commands loaded for the module
        self.commands = {}
        # module name -> list of event_name/handler pairs loaded for the module
        self.event_handlers = {}
        # module name -> dict of tab types; tab type -> commands
        # loaded by the module
        self.tab_commands = {}
        # module name → dict of keys/handlers loaded for the module
        self.keys = {}
        # module name → dict of tab types; tab type → list of keybinds (tuples)
        self.tab_keys = {}

        self.finder = machinery.PathFinder()

        self.initial_set_plugins_dir()
        self.initial_set_plugins_conf_dir()
        self.fill_load_path()

    def disable_plugins(self) -> None:
        for plugin in set(self.plugins.keys()):
            self.unload(plugin, notify=False)

    def set_rdeps(self, name: str) -> None:
        """
        Runs through plugin dependencies to build the reverse dependencies table.
        """

        if name not in self.rdeps:
            self.rdeps[name] = set()
        plugin = self.plugins[name]
        if plugin is None:
            return
        for dep in plugin.dependencies:
            if dep not in self.rdeps:
                self.rdeps[dep] = {name}
            else:
                self.rdeps[dep].add(name)

    def load(self, name: str, notify: bool = True, unload_first: bool = True) -> None:
        """
        Load a plugin.
        """
        if not unload_first and name in self.plugins:
            return
        if name in self.plugins:
            self.unload(name)
        module = self._find_module(name, notify=notify)
        if not module:
            return
        self._init_module(name, module, notify=notify)

    def _find_module(self, name: str, notify: bool = True) -> ModuleType | None:
        module = None
        try:
            spec = self.finder.find_spec(name, self.load_path)
            if spec and spec.loader:
                module = util.module_from_spec(spec)
                if module is None:
                    log.debug('Failed to load plugin %s from loader', name)
                else:
                    log.debug(
                        'Found candidate loader for plugin %s: %r',
                        name, spec.loader,
                    )
                    spec.loader.exec_module(module)
            else:
                entrypoints = entry_points(group='poezio_plugins')
                to_import = f'poezio_plugins.{name}'
                entrypoint_found = entrypoints.select(name=name)
                if entrypoint_found:
                    to_import = entrypoints[name].value
                try:
                    module = import_module(to_import)
                except Exception as exn:
                    log.debug('Failed to import plugin: %s\n%r', name,
                              exn, exc_info=True)
            if not module:
                if notify:
                    self.core.information(f'Could not find plugin: {name}',
                                          'Error')
                return None
            log.debug('Plugin %s loaded from "%s"', name, module.__file__)
        except Exception as e:
            log.debug("Could not load plugin %s", name, exc_info=True)
            if notify:
                self.core.information(f"Could not load plugin {name}: {e}",
                                      'Error')
        return module

    def _init_module(self, name: str, module: ModuleType, notify: bool = True) -> None:
        """
        Init the plugin once the module has been found.
        Fails to load if dependencies do not load.
        """
        self.modules[name] = module
        self.commands[name] = {}
        self.keys[name] = {}
        self.tab_keys[name] = {}
        self.tab_commands[name] = {}
        self.event_handlers[name] = []
        self.plugins[name] = None
        for dep in module.Plugin.dependencies:
            self.load(dep, unload_first=False)
            if dep not in self.plugins:
                if notify:
                    self.core.information(
                        f'Plugin {name} could not load because of dependency {dep}',
                        'Error',
                    )
                self.unload(name, notify=False)
                return
            # Add reference of the dep to the plugin's usage
            module.Plugin.refs[dep] = self.plugins[dep]

        try:
            self.plugins[name] = module.Plugin(name, PluginAPI(self.core, self, name), self.core,
                                               self.plugins_conf_dir)
        except Exception as e:
            log.exception('Error while loading the plugin %s', name)
            self.unload(name, notify=False)
            if notify:
                self.core.information(
                    f'Unable to load the plugin {name}: {e}', 'Error')
        else:
            self.set_rdeps(name)
            if notify:
                self.core.information(f'Plugin {name} loaded', 'Info')

    def unload(self, name: str, notify: bool = True) -> None:
        """
        Unloads plugin as well as plugins depending on it.
        """

        if name in self.plugins:
            try:
                plugin = self.plugins[name]
                if plugin is not None:
                    plugin._unloading = True  # Prevents loops
                    for rdep in self.rdeps[name].copy():
                        if rdep in self.plugins:
                            rdep_plugin = self.plugins[rdep]
                            if rdep_plugin and not rdep_plugin._unloading:
                                self.unload(rdep)
                                if rdep in self.plugins:
                                    log.debug('Failed to unload reverse dependency %s first.', rdep)
                                    return

                for command in self.commands[name].keys():
                    del self.core.commands[command]
                for key in self.keys[name].keys():
                    del self.core.key_func[key]
                for tab in list(self.tab_commands[name].keys()):
                    for tab_command in self.tab_commands[name][tab][:]:
                        self.del_tab_command(name, getattr(tabs, tab),
                                             tab_command[0])
                    del self.tab_commands[name][tab]
                for tab in list(self.tab_keys[name].keys()):
                    for tab_key in self.tab_keys[name][tab][:]:
                        self.del_tab_key(name, getattr(tabs, tab), tab_key[0])
                    del self.tab_keys[name][tab]
                for event_name, handler in self.event_handlers[name][:]:
                    self.del_event_handler(name, event_name, handler)

                plugin = self.plugins[name]
                if plugin is not None:
                    plugin.unload()
                del self.plugins[name]
                del self.rdeps[name]
                del self.commands[name]
                del self.keys[name]
                del self.tab_commands[name]
                del self.event_handlers[name]
                if notify:
                    self.core.information('Plugin %s unloaded' % name, 'Info')
            except Exception as e:
                log.debug("Could not unload plugin %s", name, exc_info=True)
                self.core.information(
                    "Could not unload plugin %s: %s" % (name, e), 'Error')

    def add_command(self,
                    module_name: str,
                    name: str,
                    handler: Callable[[str], Any],
                    help: str,
                    completion: GenericCompletion = None,
                    short: str = '',
                    usage: str = '') -> None:
        """
        Add a global command.
        """
        if name in self.core.commands:
            raise KeyError("Command '%s' already exists" % (name, ))

        commands = self.commands[module_name]
        commands[name] = Command(handler, help, completion, short, usage)
        self.core.commands[name] = commands[name]

    def del_command(self, module_name: str, name: str) -> None:
        """
        Remove a global command added through add_command.
        """
        if name in self.commands[module_name]:
            del self.commands[module_name][name]
            if name in self.core.commands:
                del self.core.commands[name]

    def add_tab_command(self,
                        module_name: str,
                        tab_type: Type[tabs.Tab],
                        name: str,
                        handler: Callable[[str], Any],
                        help: str,
                        completion: GenericCompletion = None,
                        short: str = '',
                        usage: str = '') -> None:
        """
        Add a command only for a type of Tab.
        """
        commands = self.tab_commands[module_name]
        t = tab_type.__name__
        if name in tab_type.plugin_commands:
            return
        if t not in commands:
            commands[t] = []
        commands[t].append((name, handler, help, completion))
        tab_type.plugin_commands[name] = Command(handler, help, completion,
                                                 short, usage)
        for tab in self.core.tabs:
            if isinstance(tab, tab_type):
                tab.update_commands()

    def del_tab_command(self, module_name: str, tab_type: Type[tabs.Tab],
                        name: str) -> None:
        """
        Remove a command added through add_tab_command.
        """
        commands = self.tab_commands[module_name]
        t = tab_type.__name__
        if t not in commands:
            return
        for command in commands[t]:
            if command[0] == name:
                commands[t].remove(command)
                del tab_type.plugin_commands[name]
                for tab in self.core.tabs:
                    if isinstance(tab, tab_type) and name in tab.commands:
                        del tab.commands[name]

    def add_tab_key(self, module_name: str, tab_type: Type[tabs.Tab], key: str,
                    handler: Callable[[], Any]) -> None:
        """
        Associate a key binding to a handler only for a type of Tab.
        """
        keys = self.tab_keys[module_name]
        t = tab_type.__name__
        if key in tab_type.plugin_keys:
            return
        if t not in keys:
            keys[t] = []
        keys[t].append((key, handler))
        tab_type.plugin_keys[key] = handler
        for tab in self.core.tabs:
            if isinstance(tab, tab_type):
                tab.update_keys()

    def del_tab_key(self, module_name: str, tab_type: Type[tabs.Tab], key: str) -> None:
        """
        Remove a key binding added through add_tab_key.
        """
        keys = self.tab_keys[module_name]
        t = tab_type.__name__
        if t not in keys:
            return
        for _key in keys[t]:
            if _key[0] == key:
                keys[t].remove(_key)
                del tab_type.plugin_keys[key]
                for tab in self.core.tabs:
                    if isinstance(tab, tab_type) and key in tab.key_func:
                        del tab.key_func[key]

    def add_key(self, module_name: str, key: str, handler: Callable[..., Any]) -> None:
        """
        Associate a global key binding to a handler, except if it
        already exists.
        """
        if key in self.core.key_func:
            raise KeyError("Key '%s' already exists" % (key, ))
        keys = self.keys[module_name]
        keys[key] = handler
        self.core.key_func[key] = handler

    def del_key(self, module_name: str, key: str) -> None:
        """
        Remove a global key binding added by a plugin.
        """
        if key in self.keys[module_name]:
            del self.keys[module_name][key]
            if key in self.core.key_func:
                del self.core.commands[key]

    def add_event_handler(self, module_name: str, event_name: str,
                          handler: Callable[..., Any],
                          priority: int = 50) -> None:
        """
        Add an event handler. If event_name isn’t in the event list, assume
        it is a slixmpp event.
        """
        eh = self.event_handlers[module_name]
        eh.append((event_name, handler))
        if event_name in self.core.events.events:
            self.core.events.add_event_handler(event_name, handler,
                                               priority=priority)
        else:
            self.core.xmpp.add_event_handler(event_name, handler)

    def del_event_handler(self, module_name: str, event_name: str,
                          handler: Callable[..., Any]) -> None:
        """
        Remove an event handler if it exists.
        """
        if event_name in self.core.events.events:
            self.core.events.del_event_handler(None, handler)
        else:
            self.core.xmpp.del_event_handler(event_name, handler)
        handlers = self.event_handlers[module_name]
        self.event_handlers[module_name] = [
            current_handler for current_handler in handlers
            if current_handler != (event_name, handler)
        ]

    def completion_load(self, the_input: 'Input') -> Completion:
        """
        completion function that completes the name of the plugins, from
        all .py files in plugins_dir
        """
        names = set()
        for path_ in self.load_path:
            try:
                add = set(os.listdir(path_))
                names |= add
            except OSError:
                pass
        plugins_files = [
            name[:-3] for name in names if name.endswith('.py')
            and name != '__init__.py' and not name.startswith('.')
        ]
        plugins_files.sort()
        position = the_input.get_argument_position(quoted=False)
        return Completion(
            the_input.new_completion,
            plugins_files,
            position,
            '',
            quotify=False)

    def completion_unload(self, the_input: 'Input') -> Completion:
        """
        completion function that completes the name of loaded plugins
        """
        position = the_input.get_argument_position(quoted=False)
        return Completion(
            the_input.new_completion,
            sorted(self.plugins.keys()),
            position,
            '',
            quotify=False)

    def on_plugins_dir_change(self, _: Any, new_value: str) -> None:
        self.plugins_dir = Path(new_value).expanduser()
        self.check_create_plugins_dir()
        self.fill_load_path()

    def on_plugins_conf_dir_change(self, _: Any, new_value: str) -> None:
        self.plugins_conf_dir = Path(new_value).expanduser()
        self.check_create_plugins_conf_dir()

    def initial_set_plugins_conf_dir(self) -> None:
        """
        Create the plugins_conf_dir
        """
        plugins_conf_dir = config.getstr('plugins_conf_dir')
        self.plugins_conf_dir = Path(plugins_conf_dir).expanduser(
        ) if plugins_conf_dir else xdg.CONFIG_HOME / 'plugins'
        self.check_create_plugins_conf_dir()

    def check_create_plugins_conf_dir(self) -> bool:
        """
        Create the plugins config directory if it does not exist.
        Returns True on success, False on failure.
        """
        if not os.access(str(self.plugins_conf_dir), os.R_OK | os.X_OK):
            try:
                self.plugins_conf_dir.mkdir(parents=True, exist_ok=True)
            except OSError:
                log.exception(
                    'Unable to create the plugin conf dir: %s',
                    self.plugins_conf_dir,
                )
                return False
        return True

    def initial_set_plugins_dir(self) -> None:
        """
        Set the plugins_dir on start
        """
        plugins_dir = config.getstr('plugins_dir')
        self.plugins_dir = Path(plugins_dir).expanduser(
        ) if plugins_dir else xdg.DATA_HOME / 'plugins'
        self.check_create_plugins_dir()

    def check_create_plugins_dir(self) -> bool:
        """
        Create the plugins directory if it does not exist.
        Returns True on success, False on failure.
        """
        if not os.access(str(self.plugins_dir), os.R_OK | os.X_OK):
            try:
                self.plugins_dir.mkdir(parents=True, exist_ok=True)
            except OSError:
                log.exception(
                    'Unable to create the plugins dir: %s',
                    self.plugins_dir,
                )
                return False
        return True

    def fill_load_path(self) -> None:
        """
        Append the global packages and the source directory if available
        """

        self.load_path: list[str] = []

        default_plugin_path = path.join(
            path.dirname(path.dirname(__file__)), 'plugins')

        if os.access(default_plugin_path, os.R_OK | os.X_OK):
            self.load_path.insert(0, default_plugin_path)

        if os.access(str(self.plugins_dir), os.R_OK | os.X_OK):
            self.load_path.append(str(self.plugins_dir))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2019 Maxime “pep” Buquet <pep@bouah.net>
#
# Distributed under terms of the GPL-3.0+ license.

"""
Usage
-----

Base64 encryption plugin.

This plugin also respects security guidelines listed in XEP-0419.

.. glossary::
    /b64
        **Usage:** ``/b64``

        This command enables encryption of outgoing messages for the current
        tab.
"""

from base64 import b64decode, b64encode
from typing import List, Optional
from slixmpp import Message, JID

from poezio.plugin_e2ee import E2EEPlugin
from poezio.tabs import (
    ChatTab,
    MucTab,
    PrivateTab,
    DynamicConversationTab,
    StaticConversationTab,
)


class Plugin(E2EEPlugin):
    """Base64 Plugin"""

    encryption_name = 'base64'
    encryption_short_name = 'b64'
    eme_ns = 'urn:xmpps:base64:0'

    # This encryption mechanism is using <body/> as a container
    replace_body_with_eme = False

    # In what tab is it ok to use this plugin. Here we want all of them
    supported_tab_types = (
        MucTab,
        PrivateTab,
        DynamicConversationTab,
        StaticConversationTab,
    )

    async def decrypt(self, message: Message, jid: Optional[JID], _tab: Optional[ChatTab]) -> None:
        """
            Decrypt base64
        """
        body = message['body']
        message['body'] = b64decode(body.encode()).decode()

    async def encrypt(self, message: Message, _jid: Optional[List[JID]], _tab: ChatTab) -> None:
        """
            Encrypt to base64
        """
        # TODO: Stop using <body/> for this. Put the encoded payload in another element.
        body = message['body']
        message['body'] = b64encode(body.encode()).decode()



# type: ignore
# Typing for this file is better left untold
"""
A module that monkey patches the standard asyncio module to add an
idle_call() method to the main loop. This method is used to execute a
callback whenever the loop is not busy handling anything else. This means
that it is a callback with lower priority than IO, timer, or even
call_soon() ones. These callback are called only once each.
"""

import asyncio
import functools
import collections
from asyncio import events
from inspect import iscoroutinefunction
from typing import Callable, Any

import slixmpp


class BadIdleCallParam(TypeError):
    def __init__(self) -> None:
        super().__init__("coroutines cannot be used with idle_call()")


def monkey_patch_asyncio_slixmpp() -> None:
    def idle_call(self: asyncio.AbstractEventLoop, callback: Callable[[], None]) -> None:
        if iscoroutinefunction(callback):
            raise BadIdleCallParam()
        handle = events.Handle(callback, [], self)
        self._idle.append(handle)

    def my_run_once(self: asyncio.AbstractEventLoop) -> None:
        if self._idle:
            self._ready.append(events.Handle(lambda: None, (), self))
        real_run_once(self)
        if self._idle:
            handle = self._idle.popleft()
            handle._run()

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None
    if loop is None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    cls = loop.__class__
    cls._idle = collections.deque()
    cls.idle_call = idle_call
    real_run_once = cls._run_once
    cls._run_once = my_run_once

    spawn_event = slixmpp.xmlstream.XMLStream._spawn_event

    def patchy(self: slixmpp.xmlstream.XMLStream, xml: Any) -> None:
        self.loop.idle_call(functools.partial(spawn_event, self, xml))

    slixmpp.xmlstream.XMLStream._spawn_event = patchy



"""
Windows relevant for the listing tabs, not much else
"""

import logging
import curses
from typing import Sequence

from poezio.windows.base_wins import Win
from poezio.theming import to_curses_attr, get_theme

log = logging.getLogger(__name__)


class ListWin(Win):
    """
    A list (with no depth, so not for the roster) that can be
    scrolled up and down, with one selected line at a time
    """

    __slots__ = ('_columns', '_columns_sizes', 'sorted_by', 'lines',
                 '_selected_row', '_starting_pos')

    def __init__(self, columns: dict[str, int], with_headers: bool = True) -> None:
        Win.__init__(self)
        self._columns: dict[str, int] = columns
        self._columns_sizes: dict[str, int] = {}
        self.sorted_by = (None, None)  # for example: ('name', '↑')
        self.lines: list[tuple[str, ...]] = []
        self._selected_row = 0
        self._starting_pos = 0  # The column number from which we start the refresh

    @property
    def pos(self) -> int:
        if len(self.lines) > self.height:
            return len(self.lines)
        else:
            return 0

    def empty(self) -> None:
        """
        empty the list and reset some important values as well
        """
        self.lines = []
        self._selected_row = 0
        self._starting_pos = 0

    def resize_columns(self, dic: dict[str, int]) -> None:
        """
        Resize the width of the columns
        """
        self._columns_sizes = dic

    def sort_by_column(self, col_name: str, asc: bool = True) -> None:
        """
        Sort the list by the given column, ascendant or descendant
        """
        if not col_name:
            return
        elif asc:
            self.lines.sort(key=lambda x: x[self._columns[col_name]])
        else:
            self.lines.sort(
                key=lambda x: x[self._columns[col_name]], reverse=True)
        self.refresh()
        curses.doupdate()

    def add_lines(self, lines: list[tuple[str, ...]]) -> None:
        """
        Append some lines at the end of the list
        """
        if not lines:
            return
        self.lines.extend(lines)

    def set_lines(self, lines: Sequence[tuple[str, ...]]) -> None:
        """
        Set the lines to another list
        """
        if not lines:
            return
        self.lines = list(lines)

    def get_selected_row(self) -> tuple[str, ...] | None:
        """
        Return the tuple representing the selected row
        """
        if self._selected_row is not None and self.lines:
            return self.lines[self._selected_row]
        return None

    def refresh(self) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        lines = self.lines[self._starting_pos:self._starting_pos + self.height]
        color = to_curses_attr(get_theme().COLOR_INFORMATION_BAR)
        for y, line in enumerate(lines):
            x = 0
            for col in self._columns.items():
                try:
                    txt = line[col[1]] or ''
                except KeyError:
                    txt = ''
                size = self._columns_sizes[col[0]]
                txt += ' ' * (size - len(txt))
                if not txt:
                    continue
                if line is self.lines[self._selected_row]:
                    self.addstr(y, x, txt[:size], color)
                else:
                    self.addstr(y, x, txt[:size])
                x += size
        self._refresh()

    def move_cursor_down(self) -> bool:
        """
        Move the cursor Down
        """
        if not self.lines:
            return False
        if self._selected_row < len(self.lines) - 1:
            self._selected_row += 1
        while self._selected_row >= self._starting_pos + self.height:
            self._starting_pos += self.height // 2
        if self._starting_pos < 0:
            self._starting_pos = 0
        return True

    def move_cursor_up(self) -> bool:
        """
        Move the cursor Up
        """
        if not self.lines:
            return False
        if self._selected_row > 0:
            self._selected_row -= 1
        while self._selected_row < self._starting_pos:
            self._starting_pos -= self.height // 2
        return True

    def scroll_down(self) -> bool:
        if not self.lines:
            return False
        self._selected_row += self.height
        if self._selected_row > len(self.lines) - 1:
            self._selected_row = len(self.lines) - 1
        while self._selected_row >= self._starting_pos + self.height:
            self._starting_pos += self.height // 2
        if self._starting_pos < 0:
            self._starting_pos = 0
        return True

    def scroll_up(self) -> bool:
        if not self.lines:
            return False
        self._selected_row -= self.height + 1
        if self._selected_row < 0:
            self._selected_row = 0
        while self._selected_row < self._starting_pos:
            self._starting_pos -= self.height // 2
        return True


class ColumnHeaderWin(Win):
    """
    A class displaying the column's names
    """

    __slots__ = ('_columns', '_columns_sizes', '_column_sel', '_column_order',
                 '_column_order_asc')

    def __init__(self, columns: list[str]) -> None:
        Win.__init__(self)
        self._columns = columns
        self._columns_sizes: dict[str, int] = {}
        self._column_sel = ''
        self._column_order = ''
        self._column_order_asc = False

    def resize_columns(self, dic: dict[str, int]) -> None:
        self._columns_sizes = dic

    def get_columns(self) -> list[str]:
        return self._columns

    def refresh(self) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        x = 0
        theme = get_theme()
        for col in self._columns:
            txt = col
            if col in self._column_order:
                if self._column_order_asc:
                    txt += theme.CHAR_COLUMN_ASC
                else:
                    txt += theme.CHAR_COLUMN_DESC
            # ⇓⇑↑↓⇧⇩▲▼
            size = self._columns_sizes[col]
            txt += ' ' * (size - len(txt))
            if col in self._column_sel:
                self.addstr(
                    0, x, txt,
                    to_curses_attr(theme.COLOR_COLUMN_HEADER_SEL))
            else:
                self.addstr(0, x, txt,
                            to_curses_attr(theme.COLOR_COLUMN_HEADER))
            x += size
        self._refresh()

    def sel_column(self, dic: str) -> None:
        self._column_sel = dic

    def get_sel_column(self) -> str:
        return self._column_sel

    def set_order(self, order: bool) -> None:
        self._column_order = self._column_sel
        self._column_order_asc = order

    def get_order(self) -> bool:
        if self._column_sel == self._column_order:
            return self._column_order_asc
        else:
            return False

    def sel_column_left(self) -> None:
        if self._column_sel in self._columns:
            index = self._columns.index(self._column_sel)
            if index > 1:
                index = index - 1
            else:
                index = 0
        else:
            index = 0
        self._column_sel = self._columns[index]
        self.refresh()

    def sel_column_right(self) -> None:
        if self._column_sel in self._columns:
            index = self._columns.index(self._column_sel)
            if index < len(self._columns) - 2:
                index = index + 1
            else:
                index = len(self._columns) - 1
        else:
            index = len(self._columns) - 1
        self._column_sel = self._columns[index]
        self.refresh()



"""
This plugin lets you execute a system command through poezio.

Usage
-----

.. warning:: Running commands that start a daemon or an interface is not a good
             idea.

.. glossary::

    /exec
        **Usage:** ``/exec [-o|-O] <command>``

        Execute a system command.

        ::

            /exec command

        Will give you the result in the information buffer.

        ::

            /exec -o command

        Will send the result of the command into the current tab, if possible.

        ::

            /exec -O command

        Will send the result of the command and the command summary into the current
        tab, if possible.

"""

from poezio.plugin import BasePlugin
from poezio import common
import asyncio
import subprocess


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'exec',
            self.command_exec,
            usage='[-o|-O] <command>',
            help=(
                'Execute a shell command and prints the result in the '
                'information buffer. The command should be ONE argument, '
                'that means it should be between \"\". The first argument '
                '(before the command) can be -o or -O. If -o is specified, '
                'it sends the result in the current conversation. If -O is '
                'specified, it sends the command and its result in the '
                'current conversation.\nExample: /exec -O \"uptime\" will '
                'send “uptime\n20:36:19 up  3:47,  4 users,  load average: '
                '0.09, 0.13, 0.09” in the current conversation.'
            ),
            short='Execute a command')

    async def async_exec(self, command: str, arg: str | None) -> None:
        create = asyncio.create_subprocess_exec(
            'sh',
            '-c',
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        try:
            process = await create
        except OSError as e:
            self.api.information('Failed to execute command: %s' % (e, ),
                                 'Error')
            return
        raw_stdout, raw_stderr = await process.communicate()
        result = raw_stdout.decode('utf-8')
        stderr = raw_stderr.decode('utf-8')
        if arg == '-o':
            if not self.api.send_message('%s' % (result, )):
                self.api.information(
                    'Cannot send result (%s), this is not a conversation tab' %
                    result, 'Error')
        elif arg == '-O':
            if not self.api.send_message('%s:\n%s' % (command, result)):
                self.api.information(
                    'Cannot send result (%s), this is not a conversation tab' %
                    result, 'Error')
        else:
            self.api.information('%s:\n%s' % (command, result), 'Info')
        if stderr:
            self.api.information('stderr for %s:\n%s' % (command, stderr),
                                 'Info')
        await process.wait()

    def command_exec(self, args: str) -> None:
        list_args = common.shell_split(args)
        if len(list_args) == 1:
            command = list_args[0]
            arg = None
        elif len(list_args) == 2:
            command = list_args[1]
            arg = args[0]
        else:
            self.api.run_command('/help exec')
            return
        self.core.register_task(self.async_exec(command, arg))



"""
Define the TextBuffer class

A text buffer contains a list of intermediate representations of messages
(not xml stanzas, but neither the Lines used in windows.py.

Each text buffer can be linked to multiple windows, that will be rendered
independently by their TextWins.
"""

from __future__ import annotations

import logging

from typing import (
    Type,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
    Sequence,
)
from dataclasses import dataclass
from datetime import datetime
from poezio.config import config
from poezio.ui.types import (
    BaseMessage,
    Message,
    MucOwnJoinMessage,
    MucOwnLeaveMessage,
    DeletedMessage,
)

if TYPE_CHECKING:
    from poezio.windows.text_win import TextWin
    from poezio.user import User
    from slixmpp import JID


log = logging.getLogger(__name__)


class CorrectionError(Exception):
    """
    Exception raised when the text buffer fails to correct a message.
    There are plenty of reasons for that.
    """
    def __init__(self, found: bool = True, is_message: bool = True,  # noqa: B042
                 same_user: bool = True, delayed: bool = False,
                 no_identity: bool = False, no_match: bool = False,
                 ):
        if not found:
            msg = 'Message not found in text buffer.'
        elif not is_message:
            msg = 'Message in text buffer is not a proper message.'
        elif not same_user:
            msg = 'Message and corrections are from different users.'
        elif delayed:
            msg = 'Cannot correct a delayed message.'
        elif no_identity:
            msg = 'Could not verify the sender’s identity.'
        elif no_match:
            msg = (
                'Messages and corrections have not been'
                'sent by the same fullJID.'
            )
        else:
            msg = 'unexplained correction error'

        super().__init__(msg)


class DeletionError(Exception):
    """
    Error raised when trying to delete a message from the text buffer and
    failing the permission check (either similar user, similar jid, or
    moderated message.
    """
    def __init__(self,  # noqa: B042
                 users: tuple['User', 'User' | None] | None = None,
                 jids: tuple[str, str] | None = None) -> None:
        if users:
            msg = (
                f'Retraction failed: {users[0]} and '
                f'{users[1]} are different users!'
            )
        elif jids:
            msg = (
                f'Retraction failed: {jids[0]} and '
                f'{jids[1]} are different jids!'
            )
        else:
            msg = "no jid, no user, no moderation, should not happen"
        super().__init__(msg)


class AckError(Exception):
    pass


@dataclass(slots=True)
class HistoryGap:
    """Class representing a period of non-presence inside a MUC"""
    leave_message: Optional[BaseMessage]
    join_message: Optional[BaseMessage]
    last_timestamp_before_leave: Optional[datetime]
    first_timestamp_after_join: Optional[datetime]


class TextBuffer:
    """
    This class just keep trace of messages, in a list with various
    information and attributes.
    """
    filtered_instances: tuple[Type[BaseMessage], ...]

    def __init__(self, messages_nb_limit: Optional[int] = None) -> None:

        self.filtered_instances = tuple()
        if messages_nb_limit is None:
            messages_nb_limit = config.getint('max_messages_in_memory')
        self._messages_nb_limit: int = messages_nb_limit
        # Message objects
        self.messages: List[BaseMessage] = []
        # COMPAT: Correction id -> Original message id.
        self.correction_ids: Dict[str, str] = {}
        # we keep track of one or more windows
        # so we can pass the new messages to them, as they are added, so
        # they (the windows) can build the lines from the new message
        self._windows: List[TextWin] = []

    def add_window(self, win: TextWin) -> None:
        self._windows.append(win)

    def filter_classes(self, filtered: tuple[Type[BaseMessage], ...]) -> bool:
        """
        Filter out specific types of messages from the windows (i.e. the real
        display). Those messages are still in the text buffer and will be
        displayed again when the filter is removed.

        note: it is costly to change the filter
        """
        old_filter = self.filtered_instances
        self.filtered_instances = filtered
        if old_filter != filtered:
            for window in self._windows:
                window.rebuild_everything(self)
            return True
        return False

    def filter_messages(self) -> list[BaseMessage]:
        """Return a list of messages, filtered if a filter was defined."""
        if not self.filtered_instances:
            return self.messages
        else:
            return [
                message for message in self.messages
                if not isinstance(message, self.filtered_instances)
            ]

    def find_last_gap_muc(self) -> Optional[HistoryGap]:
        """
        Find the last known history gap contained in buffer.

        The idea is to be able to fetch all messages between a leave
        and a join. It is made quite complicated by the fact that due to
        possible network issues on our side, on the remote side, or on our
        server side (or other bugs in poezio!), we can disappear from a room
        without leaving properly.

        We want to know the last timestamp of the message before we left (in
        case a race condition made us miss something) if any, and the first
        timestamp of the message after our join if any, so that the mam query
        can be properly calibrated.

        """
        leave: Optional[Tuple[int, BaseMessage]] = None
        join: Optional[Tuple[int, BaseMessage]] = None
        for i, item in enumerate(reversed(self.messages)):
            if isinstance(item, MucOwnLeaveMessage):
                leave = (len(self.messages) - i - 1, item)
                break
            elif join and isinstance(item, MucOwnJoinMessage):
                leave = (len(self.messages) - i - 1, item)
                break
            if isinstance(item, MucOwnJoinMessage):
                join = (len(self.messages) - i - 1, item)

        last_timestamp = None
        first_timestamp = datetime.now()

        # Identify the special case when we got disconnected from a chatroom
        # without receiving or sending the relevant presence, therefore only
        # having two joins with no leave, and messages in the middle.
        if leave and join and isinstance(leave[1], MucOwnJoinMessage):
            for i in range(join[0] - 1, leave[0], - 1):
                if isinstance(self.messages[i], Message):
                    leave = (
                        i,
                        self.messages[i]
                    )
                    last_timestamp = self.messages[i].time
                    break
        # If we have a normal gap but messages inbetween, it probably
        # already has history, so abort there without returning it.
        if join and leave:
            for i in range(leave[0] + 1, join[0], 1):
                if isinstance(self.messages[i], Message):
                    return None
        elif not (join or leave):
            return None

        # If a leave message is found, get the last Message timestamp
        # before it.
        if leave is None:
            leave_msg = None
        elif last_timestamp is None:
            leave_msg = leave[1]
            for i in range(leave[0], 0, -1):
                if isinstance(self.messages[i], Message):
                    last_timestamp = self.messages[i].time
                    break
                # If we encounter a join, assume that we can take the "leave"
                # time instead.
                if isinstance(self.messages[i], MucOwnJoinMessage):
                    break
            if not last_timestamp:
                last_timestamp = leave[1].time
        else:
            leave_msg = leave[1]
        # If a join message is found, get the first Message timestamp
        # after it, or the current time.
        if join is None:
            join_msg = None
        else:
            join_msg = join[1]
            for i in range(join[0], len(self.messages)):
                msg = self.messages[i]
                if isinstance(msg, Message) and msg.time < first_timestamp:
                    first_timestamp = msg.time
                    break
        return HistoryGap(
            leave_message=leave_msg,
            join_message=join_msg,
            last_timestamp_before_leave=last_timestamp,
            first_timestamp_after_join=first_timestamp,
        )

    def get_gap_index(self, gap: HistoryGap) -> Optional[int]:
        """Find the first index to insert into inside a gap"""
        if gap.leave_message is None:
            return 0
        for i, msg in enumerate(self.messages):
            if msg is gap.leave_message:
                return i + 1
        return None

    def add_history_messages(self, messages: Sequence[BaseMessage],
                             gap: Optional[HistoryGap] = None) -> None:
        """Insert history messages at their correct place """
        index = 0
        new_index = None
        if gap is not None:
            new_index = self.get_gap_index(gap)
            if new_index is None:  # Not sure what happened, abort
                return
            index = new_index
        # Crappy heuristic: assume if a message has the same text and nickname
        # as the last message we know, we only need to take the messages that
        # come afterwards.
        if gap and isinstance(gap.leave_message, Message):
            leave_msg = gap.leave_message
            for i, message in enumerate(messages[:]):
                if not isinstance(message, Message):
                    continue
                if (message.txt == leave_msg.txt
                        and message.nickname == leave_msg.nickname):
                    messages = messages[i+1:]
                    break

        for message in messages:
            self.messages.insert(index, message)
            index += 1
        for window in self._windows:  # make the associated windows
            window.rebuild_everything(self)

    @property
    def last_message(self) -> Optional[BaseMessage]:
        return self.messages[-1] if self.messages else None

    def add_message(self, msg: BaseMessage) -> int:
        """
        Create a message and add it to the text buffer
        """
        self.messages.append(msg)

        while len(self.messages) > self._messages_nb_limit:
            self.messages.pop(0)

        if isinstance(msg, self.filtered_instances):
            return 0

        ret_val = 0
        show_timestamps = config.getbool('show_timestamps')
        nick_size = config.getbool('max_nick_length')
        for window in self._windows:  # make the associated windows
            # build the lines from the new message
            nb = window.build_new_message(
                msg,
                timestamp=show_timestamps,
                nick_size=nick_size)
            if ret_val == 0:
                ret_val = nb
            if window.pos != 0:
                window.scroll_up(nb)

        return min(ret_val, 1)

    def _find_message(self, orig_id: str) -> Tuple[str, int]:
        """
        Find a message in the text buffer from its message id
        """
        # When looking for a message, ensure the id doesn't appear in a
        # message we've removed from our message list. If so return the index
        # of the corresponding id for the original message instead.
        orig_id = self.correction_ids.get(orig_id, orig_id)

        for i in range(len(self.messages) - 1, -1, -1):
            msg = self.messages[i]
            if msg.identifier == orig_id:
                return (orig_id, i)
        return (orig_id, -1)

    def _find_message_stanza_id(self, orig_id: str) -> Tuple[str, int]:
        """
        Find a message in the text buffer from its message id
        """
        # When looking for a message, ensure the id doesn't appear in a
        # message we've removed from our message list. If so return the index
        # of the corresponding id for the original message instead.
        orig_id = self.correction_ids.get(orig_id, orig_id)

        for i in range(len(self.messages) - 1, -1, -1):
            msg = self.messages[i]
            if hasattr(msg, 'stanza_id'):
                if msg.stanza_id == orig_id:
                    return (orig_id, i)
        return (orig_id, -1)

    def react(self, to: str, by: str, reactions: set[str],
              use_stanza_ids: bool = False) -> bool:
        """
        Add or subtract reactions to/from a message.

        :param to: Identifier of the message
        :param by: An identifier relevant to the current context
        :param reactions: set of reactions to for the entity related to "by"
        :param use_stanza_ids: Whether to use stanza ids (e.g. in a MUC)
        :returns: True if the reactions sets were changed
        """
        if use_stanza_ids:
            _, pos = self._find_message_stanza_id(to)
        else:
            _, pos = self._find_message(to)
        if pos == -1:
            log.debug('Unable to find message %s to add reactions', to)
            return False
        message = self.messages[pos]
        if not isinstance(message, Message):
            log.debug('Message %s is not a Message, should not happen', message)
            return False
        if reactions and not message.reactions:
            message.reactions = {}
        added_reactions = set()
        changed = False
        # Appease mypy
        if message.reactions is None:
            return False
        # Update existing id sets
        for emoji, ids in list(message.reactions.items()):
            if emoji in reactions:
                if by not in ids:
                    changed |= True
                    ids.add(by)
                added_reactions.add(emoji)
            elif by in ids:
                ids.discard(by)
                changed |= True
                # Remove set when empty
                if not ids:
                    del message.reactions[emoji]
        # Reaction is the first of its kind
        for emoji in (reactions - added_reactions):
            message.reactions[emoji] = {by}
            changed |= True

        # Force message refresh
        if changed:
            for window in self._windows:
                window.modify_message(message.identifier, message=message)
        return changed

    def ack_message(self, old_id: str, jid: JID) -> Union[None, bool, Message]:
        """Mark a message as acked"""
        return self._edit_ack(1, old_id, jid)

    def nack_message(self, error: str, old_id: str,
                     jid: JID) -> Union[None, bool, Message]:
        """Mark a message as errored"""
        return self._edit_ack(-1, old_id, jid, append=error)

    def _edit_ack(self, value: int, old_id: str, jid: JID,
                  append: str = '') -> Union[None, bool, Message]:
        """
        Edit the ack status of a message, and optionally
        append some text.
        """
        _, i = self._find_message(old_id)
        if i == -1:
            return None
        msg = self.messages[i]
        if not isinstance(msg, Message):
            return None
        if msg.ack == 1:  # Message was already acked
            return False
        if msg.jid is None or msg.jid.bare != jid.bare:
            raise AckError('Wrong JID for message id %s (was %s, expected %s)'
                           % (old_id, msg.jid, jid))

        msg.ack = value
        if append:
            msg.txt += append
        # The message itself has not changed but we modify it in the buffer
        # to display the ack element without a full rebuild
        for window in self._windows:
            window.modify_message(old_id, msg)
        return msg

    def delete_message(self, orig_id: str, *,
                       jid: Optional[JID] = None,
                       user: Optional[User] = None,
                       muc: bool = False,
                       moderation: bool = False,
                       tombstone: Optional[str] = None) -> bool:
        """
        Remove a message from the buffer, either due to XEP-0424 or XEP-0425.
        0425 sets the moderation flag, bypassing other checks, while
        otherwise MUC user or bare jid has to match.

        In the MUC case we need to search by stanza-id, but in the other cases
        we need to search by message id.

        Optionally add a tombstone.
        """
        if muc:
            orig_id, i = self._find_message_stanza_id(orig_id)
        else:
            orig_id, i = self._find_message(orig_id)

        if i == -1:
            log.debug(
                'Message %s not found in text_buffer, abort deletion.',
                orig_id)
            return False
        message: Message | BaseMessage = self.messages[i]
        if moderation:
            # Nothing to do, all moderation requests that come through here
            # are supposed to be valid.
            pass
        elif user and isinstance(message, Message):
            if user != message.user:
                raise DeletionError(users=(user, message.user))
        elif jid and isinstance(message, Message):
            if message.jid and jid.bare != message.jid.bare:
                raise DeletionError(jids=(jid.bare, message.jid.bare))
        else:
            raise DeletionError()
        tombstone_msg = None
        if tombstone:
            old_message: Message | None
            if isinstance(message, Message):
                old_message = message
            else:
                old_message = None

            tombstone_msg = DeletedMessage(tombstone, old_message=old_message)
            self.messages[i] = tombstone_msg
        else:
            del self.messages[i]
        for window in self._windows:
            window.modify_message(message.identifier, message=tombstone_msg)
        return True

    def modify_message(self,
                       txt: str,
                       orig_id: str,
                       new_id: str,
                       highlight: bool = False,
                       time: Optional[datetime] = None,
                       user: Optional[User] = None,
                       jid: Optional[JID] = None) -> Message:
        """
        Correct a message in a text buffer.

        Version 1.1.0 of Last Message Correction (0308) added clarifications
        that break the way poezio handles corrections. Instead of linking
        corrections to the previous correction/message as we were doing, we
        are now required to link all corrections to the original messages.
        """

        orig_id, i = self._find_message(orig_id)

        if i == -1:
            log.debug(
                'Message %s not found in text_buffer, abort replacement.',
                orig_id)
            raise CorrectionError(found=False)

        msg = self.messages[i]
        if not isinstance(msg, Message):
            raise CorrectionError(is_message=False)
        if msg.user and msg.user != user:
            raise CorrectionError(same_user=False)
        elif msg.delayed:
            raise CorrectionError(delayed=True)
        elif not msg.user and (msg.jid is None or jid is None):
            raise CorrectionError(no_identity=True)
        elif not msg.user and msg.jid != jid:
            raise CorrectionError(no_match=True)

        if not time:
            time = datetime.now()

        self.correction_ids[new_id] = orig_id
        message = Message(
            txt=txt,
            time=time,
            nickname=msg.nickname,
            nick_color=msg.nick_color,
            user=msg.user,
            identifier=orig_id,
            highlight=highlight,
            old_message=msg,
            revisions=msg.revisions + 1,
            jid=jid)
        self.messages[i] = message
        for window in self._windows:
            window.modify_message(orig_id, message)
        log.debug('Replacing message %s with %s.', orig_id, new_id)
        return message

    def del_window(self, win: TextWin) -> None:
        self._windows.remove(win)

    def find_last_message(self) -> Optional[Message]:
        """Find the last real message received in this buffer"""
        for message in reversed(self.messages):
            if isinstance(message, Message):
                return message
        return None

    def find_first_message(self) -> Optional[Message]:
        """Find the first real message received in this buffer"""
        for message in self.messages:
            if isinstance(message, Message):
                return message
        return None

    def __del__(self) -> None:
        size = len(self.messages)
        log.debug('** Deleting %s messages from textbuffer', size)



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Defines the Connection class
"""

import logging
import getpass
import subprocess
import sys
import base64
import random
import ssl
from pathlib import Path
from typing import Any, TYPE_CHECKING

import slixmpp
from slixmpp import JID, InvalidJID
from slixmpp.xmlstream import ET
from slixmpp.plugins import PluginNotFound
from slixmpp.plugins.xep_0184 import XEP_0184
from slixmpp.plugins.xep_0030 import DiscoInfo
from slixmpp.util import FileSystemCache

from poezio import common
from poezio import fixes
from poezio import xdg
from poezio.config import config

if TYPE_CHECKING:
    from poezio.core.core import Core


log = logging.getLogger(__name__)


class Connection(slixmpp.ClientXMPP):
    """
    Receives everything from Jabber and emits the
    appropriate signals
    """
    core: 'Core | None'

    def __init__(self, custom_version: str = '') -> None:
        keyfile = config.getstr('keyfile')
        certfile = config.getstr('certfile')

        device_id = config.getstr('device_id')
        if not device_id:
            rng = random.SystemRandom()
            device_id = base64.urlsafe_b64encode(
                rng.getrandbits(24).to_bytes(3, 'little')).decode('ascii')
            config.set_and_save('device_id', device_id)

        if config.getstr('jid'):
            # Field used to know if we are anonymous or not.
            # many features will be handled differently
            # depending on this setting
            self.anon = False
            jid_str = config.getstr('jid')
            password = config.getstr('password')
            eval_password = config.getstr('eval_password')
            if not password and not eval_password and not (keyfile
                                                           and certfile):
                password = getpass.getpass()
            elif not password and not (keyfile and certfile):
                sys.stderr.write(
                    "No password or certificates provided, using the eval_password command.\n"
                )
                process = subprocess.Popen(
                    ['sh', '-c', eval_password],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    close_fds=True)
                code = process.wait()
                if code != 0:
                    sys.stderr.write(
                        'The eval_password command (%s) returned a '
                        'nonzero status code: %s.\n' % (eval_password, code))
                    sys.stderr.write('Poezio will now exit\n')
                    sys.exit(code)
                if process.stdout:
                    password = process.stdout.readline().decode('utf-8').strip('\n')
                else:
                    password = ''
        else:  # anonymous auth
            self.anon = True
            jid_str = config.getstr('server')
            password = ''
        try:
            jid = JID(jid_str)
        except InvalidJID:
            sys.stderr.write('Invalid jid option: "%s" is not a valid JID\n' % jid_str)
            sys.exit(1)
        jid.resource = '%s-%s' % (
            jid.resource,
            device_id) if jid.resource else 'poezio-%s' % device_id
        # TODO: use the system language
        slixmpp.ClientXMPP.__init__(
            self, jid, password or '', lang=config.getstr('lang'))

        force_encryption = config.getbool('force_encryption')
        if force_encryption:
            self['feature_mechanisms'].unencrypted_plain = False
            self['feature_mechanisms'].unencrypted_digest = False
            self['feature_mechanisms'].unencrypted_cram = False
            self['feature_mechanisms'].unencrypted_scram = False

        self.keyfile = keyfile
        self.certfile = certfile
        if keyfile and not certfile:
            log.error(
                'keyfile is present in configuration file without certfile')
        elif certfile and not keyfile:
            log.error(
                'certfile is present in configuration file without keyfile')

        self.core = None
        self.auto_reconnect = config.getbool('auto_reconnect')
        self.auto_authorize = None
        # prosody defaults, lowest is AES128-SHA, it should be a minimum
        # for anything that came out after 2002
        self.ciphers = config.getstr(
            'ciphers', 'HIGH+kEDH:HIGH+kEECDH:HIGH:!PSK'
            ':!SRP:!3DES:!aNULL')
        self.ca_certs = None
        ca_certs = config.getlist('ca_cert_path')
        if ca_certs and ca_certs != ['']:
            self.ca_certs = list(map(Path, config.getlist('ca_cert_path')))
        interval = config.getint('whitespace_interval')
        if int(interval) > 0:
            self.whitespace_keepalive_interval = int(interval)
        else:
            self.whitespace_keepalive = False
        self.__register_slix_plugins(custom_version=custom_version)
        self.init_plugins()

    def __register_slix_plugins(self, custom_version: str = '') -> None:
        """Register all slix plugins"""
        self.register_plugin('xep_0004')
        self.register_plugin('xep_0012')
        # Must be loaded before 0030.
        self.register_plugin(
            'xep_0115', {
                'caps_node':
                'https://poez.io',
                'cache':
                FileSystemCache(
                    str(xdg.CACHE_HOME),
                    'caps',
                    encode=str,
                    decode=lambda x: DiscoInfo(ET.fromstring(x))),
            })
        self.register_plugin('xep_0030')
        self.register_plugin('xep_0045')
        self.register_plugin('xep_0048')
        self.register_plugin('xep_0050')
        self.register_plugin('xep_0054')
        self.register_plugin('xep_0060')
        self.register_plugin('xep_0066')
        self.register_plugin('xep_0070')
        self.register_plugin('xep_0071')
        self.register_plugin('xep_0077')
        self.plugin['xep_0077'].create_account = False
        self.register_plugin('xep_0084')
        self.register_plugin('xep_0085')
        self.register_plugin('xep_0153')

        # monkey-patch xep_0184 to avoid requesting receipts for messages
        # without a body
        setattr(  # noqa: B010
            XEP_0184,
            "_filter_add_receipt_request",
            fixes._filter_add_receipt_request,
        )
        self.register_plugin('xep_0184')
        self.plugin['xep_0184'].auto_ack = config.getbool('ack_message_receipts')
        self.plugin['xep_0184'].auto_request = config.getbool(
            'request_message_receipts')

        self.register_plugin('xep_0191')
        if config.getbool('enable_smacks'):
            self.register_plugin('xep_0198')
        self.register_plugin('xep_0199')

        if config.getbool('enable_user_nick'):
            self.register_plugin('xep_0172')

        if config.getbool('send_poezio_info'):
            info = {'name': 'poezio', 'version': custom_version}
            if config.getbool('send_os_info'):
                info['os'] = common.get_os_info()
            self.plugin['xep_0030'].set_identities(identities={('client',
                                                                'console',
                                                                None,
                                                                'Poezio')})
        else:
            info = {'name': '', 'version': ''}
            self.plugin['xep_0030'].set_identities(identities={('client',
                                                                'console',
                                                                None, '')})
        self.register_plugin('xep_0092', pconfig=info)
        if config.getbool('send_time'):
            self.register_plugin('xep_0202')
        self.register_plugin('xep_0203')
        self.register_plugin('xep_0224')
        self.register_plugin('xep_0231')
        self.register_plugin('xep_0249')
        self.register_plugin('xep_0257')
        self.register_plugin('xep_0280')
        self.register_plugin('xep_0297')
        self.register_plugin('xep_0308')
        self.register_plugin('xep_0313')
        self.register_plugin('xep_0334')
        self.register_plugin('xep_0352')
        self.register_plugin('xep_0359')
        try:
            self.register_plugin('xep_0363')
            self.register_plugin('xep_0454')
        except PluginNotFound:
            log.exception(
                'Failed to load HTTP File Upload plugin, it can only be '
                'used with aiohttp installed'
            )
        self.register_plugin('xep_0377')
        self.register_plugin('xep_0380')
        self.register_plugin('xep_0410')
        self.register_plugin('xep_0421')
        self.register_plugin('xep_0424')
        self.register_plugin('xep_0425')
        self.register_plugin('xep_0444')

    def set_keepalive_values(self, option: str | None = None, value: str | None = None) -> None:
        """
        Called after the XMPP session has been started, or triggered when one of
        "connection_timeout_delay" and "connection_check_interval" options
        is changed.  Unload and reload the ping plugin, with the new values.
        """
        if not self.is_connected():
            # Happens when we change the value with /set while we are not
            # connected. Do nothing in that case
            return
        ping_interval = config.getint('connection_check_interval')
        timeout_delay = config.getint('connection_timeout_delay')
        if timeout_delay <= 0:
            # We help the stupid user (with a delay of 0, poezio will try to
            # reconnect immediately because the timeout is immediately
            # passed)
            # 1 second is short, but, well
            timeout_delay = 1
        self.plugin['xep_0199'].disable_keepalive()
        # If the ping_interval is 0 or less, we just disable the keepalive
        if ping_interval > 0:
            self.plugin['xep_0199'].enable_keepalive(ping_interval,
                                                     timeout_delay)

    def get_ssl_context(self) -> ssl.SSLContext:
        """
        Override the slixmpp-provided sslcontext if the user configured
        poezio to not check certificates at all
        (e.g. when testing on localhost)
        """
        ctx = super().get_ssl_context()
        verify = config.getbool('tls_verify_cert')
        if not verify:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        return ctx

    def start(self) -> None:
        """
        Connect and process events.
        """
        custom_host = config.getstr('custom_host')
        custom_port = config.get('custom_port', 5222)
        if custom_port == -1:
            custom_port = 5222
        if custom_host:
            self.connect(host=custom_host, port=custom_port)
        elif custom_port != 5222 and custom_port != -1:
            self.connect(host=self.boundjid.host, port=custom_port)
        else:
            self.connect()

    def send_raw(self, data: str | bytes) -> None:
        """
        Overrides XMLStream.send_raw, with an event added
        """
        if self.core:
            self.core.handler.outgoing_stanza(data)  # pylint: disable=E1101
        slixmpp.ClientXMPP.send_raw(self, data)

    def _handle_disconnected(self, event: Any) -> None:
        """
        Override BaseXMPP disconnected to avoid resetting the roster whenever
        we get disconnected.
        """
        # Keep the clearing of the session bind event, though
        self.session_bind_event.clear()


class MatchAll(slixmpp.xmlstream.matcher.base.MatcherBase):
    """
    Callback to retrieve all the stanzas for the XML tab
    """

    def match(self, xml: Any) -> bool:
        "match everything"
        return True



"""
``close_all`` plugin: close all tabs except chatrooms and the contact list.

Commands
--------

.. glossary::

    /closeall
        **Usage:** ``/closeall``

        Close all tabs except the roster and chatroom tabs.
"""
from poezio.plugin import BasePlugin
from poezio import tabs
from poezio.decorators import command_args_parser


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'closeall',
            self.command_closeall,
            help='Close all non-chatroom tabs.')

    @command_args_parser.ignored
    def command_closeall(self) -> None:
        """
        /closeall
        """
        current = self.api.current_tab()
        if not isinstance(current, (tabs.RosterInfoTab, tabs.MucTab)):
            self.core.go_to_roster()

        def filter_func(x: object) -> bool:
            return not isinstance(x, (tabs.RosterInfoTab, tabs.MucTab))

        matching_tabs = list(filter(filter_func, self.core.tabs.get_tabs()))
        length = len(matching_tabs)
        for tab in matching_tabs:
            self.core.close_tab(tab)
        self.api.information('%s tabs closed.' % length, 'Info')
        self.core.refresh_window()



FORMAT_CHAR = '\x19'
# These are non-printable chars, so they should never appear in the input,
# I guess. But maybe we can find better chars that are even less risky.
FORMAT_CHARS = '\x0E\x0F\x10\x11\x12\x13\x14\x15\x16\x17\x18\x1A'



"""
Windows specific to a MUC
"""

import logging
import curses

from typing import List, Tuple, Optional

from poezio.windows.base_wins import Win

from poezio import poopt
from poezio.config import config
from poezio.theming import to_curses_attr, get_theme
from poezio.user import User

log = logging.getLogger(__name__)

CachedUser = Tuple[str, str, Optional[str], str, str]


def userlist_to_cache(userlist: List[User]) -> List[CachedUser]:
    result = []
    for user in userlist:
        result.append((user.nick, user.status, user.chatstate,
                       user.affiliation, user.role))
    return result


class UserList(Win):
    __slots__ = ('pos', 'cache')

    def __init__(self) -> None:
        Win.__init__(self)
        self.pos = 0
        self.cache: List[CachedUser] = []

    def scroll_up(self) -> bool:
        self.pos += self.height - 1
        return True

    def scroll_down(self) -> bool:
        pos = self.pos
        self.pos -= self.height - 1
        if self.pos < 0:
            self.pos = 0
        return self.pos != pos

    def draw_plus(self, y: int) -> None:
        self.addstr(y, self.width - 2, '++',
                    to_curses_attr(get_theme().COLOR_MORE_INDICATOR))

    def refresh_if_changed(self, users: List[User]) -> None:
        old = self.cache
        new = userlist_to_cache(users[self.pos:self.pos + self.height])
        if len(old) != len(new):
            self.cache = new
            self.refresh(users)
            return
        for a, b in zip(old, new):
            if a != b:
                self.cache = new
                self.refresh(users)
                return

    def refresh(self, users: List[User]) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        if config.getbool('hide_user_list'):
            return  # do not refresh if this win is hidden.
        if len(users) < self.height:
            self.pos = 0
        elif self.pos >= len(users) - self.height and self.pos != 0:
            self.pos = len(users) - self.height
        self._win.erase()
        asc_sort = (config.getstr('user_list_sort').lower() == 'asc')
        if asc_sort:
            y, _ = self._win.getmaxyx()
            y -= 1
        else:
            y = 0

        for user in users[self.pos:self.pos + self.height]:
            self.draw_role_affiliation(y, user)
            self.draw_status_chatstate(y, user)
            self.addstr(y, 2, poopt.cut_by_columns(user.nick, self.width - 2),
                        to_curses_attr(user.color))
            if asc_sort:
                y -= 1
            else:
                y += 1
            if y == self.height:
                break
        # draw indicators of position in the list
        if self.pos > 0:
            if asc_sort:
                self.draw_plus(self.height - 1)
            else:
                self.draw_plus(0)
        if self.pos + self.height < len(users):
            if asc_sort:
                self.draw_plus(0)
            else:
                self.draw_plus(self.height - 1)
        self._refresh()

    def draw_role_affiliation(self, y: int, user: User) -> None:
        theme = get_theme()
        color = theme.color_role(user.role)
        symbol = theme.char_affiliation(user.affiliation)
        self.addstr(y, 1, symbol, to_curses_attr(color))

    def draw_status_chatstate(self, y: int, user: User) -> None:
        theme = get_theme()
        show_col = theme.color_show(user.show)
        if user.chatstate == 'composing':
            char = theme.CHAR_CHATSTATE_COMPOSING
        elif user.chatstate == 'active':
            char = theme.CHAR_CHATSTATE_ACTIVE
        elif user.chatstate == 'paused':
            char = theme.CHAR_CHATSTATE_PAUSED
        else:
            char = theme.CHAR_STATUS
        self.addstr(y, 0, char, to_curses_attr(show_col))

    def resize(self, height: int, width: int, y: int, x: int) -> None:
        separator = to_curses_attr(get_theme().COLOR_VERTICAL_SEPARATOR)
        self._resize(height, width, y, x)
        self._win.attron(separator)
        self._win.vline(0, 0, curses.ACS_VLINE, self.height)
        self._win.attroff(separator)


class Topic(Win):
    __slots__ = ('_message')

    def __init__(self) -> None:
        Win.__init__(self)
        self._message = ''

    def refresh(self, topic: Optional[str] = None) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        theme = get_theme()
        self._win.erase()
        if topic is not None:
            msg = topic[:self.width - 1]
        else:
            msg = self._message[:self.width - 1]
        self.addstr(0, 0, msg, to_curses_attr(theme.COLOR_TOPIC_BAR))
        _, x = self._win.getyx()
        remaining_size = self.width - x
        if remaining_size:
            self.addnstr(' ' * remaining_size, remaining_size,
                         to_curses_attr(theme.COLOR_INFORMATION_BAR))
        self._refresh()

    def set_message(self, message: str) -> None:
        self._message = message



"""
This plugin allows sending one or more emoji reactions to other messages.

It requires ``python-emoji`` to work properly (aliases handling, emoji
validation, etc).

Command
-------

.. glossary::

    /react
        **Usage in a chat tab:** ``/react <message time> [emoji] [emoji]…``

Options
-------

.. glossary::

    complete_aliases

        **Default value:** ``false``

        If true, do not autocomplete emojis based on their real value, but
        instead use their textual names or aliases (e.g. ``:sob:`` instead of
        ``😭``).
"""

import logging

from functools import lru_cache

from poezio import common
from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from poezio.core.structs import Completion
from poezio.ui.types import Message
from poezio.windows.inputs import Input
from poezio import tabs
from typing import Literal
try:
    import emoji
except ImportError:
    emoji = None  # type: ignore


log = logging.getLogger(__name__)

DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

if emoji is None:
    log.warning(
        'python-emoji was not found, some reaction features '
        'will not be available'
    )


@lru_cache(50)
def demojize(string: str) -> str:
    """Convert an emoji to a string representation"""
    if emoji is None:
        return string
    return emoji.demojize(string)


@lru_cache(50)
def translate_str_to_emoji(string: str) -> str:
    """
    Convert a string representation to an emoji.
    Handles both normal names and aliases.
    Idempotent in all other cases.
    """
    if emoji is None:
        return string

    new_string = emoji.emojize(string)
    if new_string != string:
        return new_string
    return emoji.emojize(string, language='alias')


class Plugin(BasePlugin):
    """Message Reactions (XEP-0444) plugin"""

    default_config = {
        'default': {
            'complete_aliases': False,
        },
    }

    def init(self) -> None:
        # All chat tabs
        for _tabtype in (tabs.MucTab, tabs.DynamicConversationTab,
                         tabs.StaticConversationTab, tabs.PrivateTab):
            self.api.add_tab_command(
                _tabtype,
                name='react',
                handler=self.command_react,
                usage='<message time> [emojis]',
                help='React to a message using its timestamp (using tab complete)',
                short='React to a message',
                completion=self.completion_react,
            )

    def completion_react(self, the_input: Input) -> Completion | None:
        """Datetime and emoji completion"""
        is_muc = isinstance(self.api.current_tab(), tabs.MucTab)
        pos = the_input.get_argument_position()
        # First position: we want to complete the message timestamp
        if pos == 1:
            all_messages = self.api.get_conversation_messages()[:-100:-1]
            if is_muc:
                filtered_messages = filter(
                    lambda m: isinstance(m, Message) and m.nickname and m.stanza_id,
                    all_messages,
                )
            else:
                filtered_messages = filter(
                    lambda m: isinstance(m, Message),
                    all_messages,
                )
            messages = list(map(lambda m: m.time.strftime(DATETIME_FORMAT), filtered_messages))

            if messages:
                return Completion(the_input.auto_completion, messages, quotify=False)
        # Second position: iterate through the already received emojis
        # for this message.
        elif pos >= 2:
            complete_aliases = self.config.getbool('complete_aliases')
            time = common.shell_split(the_input.text)[1]
            found = self.api.find_message(time, DATETIME_FORMAT)
            if found and found.reactions:
                if complete_aliases:
                    to_complete = [
                        demojize(reaction) for reaction in found.reactions
                    ]
                else:
                    to_complete = list(found.reactions)
                return Completion(
                    the_input.auto_completion,
                    to_complete,
                    quotify=False,
                )
        return None

    @command_args_parser.quoted(1, 1)
    def command_react(self, args: list[str]) -> None:
        """React to a message"""
        if not args:
            return
        tab = self.api.current_tab()
        is_muc = isinstance(tab, tabs.MucTab)
        if not isinstance(tab, tabs.ChatTab):
            return

        emojis = ""
        time = args[0]
        if len(args) > 1:
            emojis = args[1]
        emoji_set = set(
            translate_str_to_emoji(emoji)
            for emoji in emojis.split() if emoji
        )

        found_msg = self.api.find_message(time, DATETIME_FORMAT)
        if not found_msg:
            self.api.information(
                f'Message with timestamp “{time}” not found.',
                'Error',
            )
            return

        mtype: Literal['chat', 'groupchat'] = 'chat'
        id_ = found_msg.identifier
        # in MUCs, use stanza-ids and "groupchat" message type
        if is_muc:
            mtype = 'groupchat'
            id_ = found_msg.stanza_id or ''
            if not id_:
                self.api.information(
                    f'Message with timestamp “{time}” '
                    'found but contained no “{urn:xmpp:sid:0}stanza-id” identifier.',
                    'Error',
                )
                return

        # We need the mfrom specifically here because we need to handle this
        # message ourselves.
        msg = self.core.xmpp.make_message(
            mtype=mtype, mto=tab.jid,
            mfrom=self.core.xmpp.boundjid,
        )
        try:
            self.core.xmpp['xep_0444'].set_reactions(msg, id_, emoji_set)
        except ValueError as exc:
            self.api.information(f'Failed setting reactions: {exc.args[0]}', 'Error')
        else:
            msg.send()
            if isinstance(tab, tabs.OneToOneTab):
                tab.handle_reactions(msg)



"""
Size Manager:
    used to check size boundaries of the whole window and
    specific tabs
"""

from poezio.windows import base_wins
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from poezio.core.core import Core

THRESHOLD_WIDTH_DEGRADE = 45
THRESHOLD_HEIGHT_DEGRADE = 10

FULL_WIDTH_DEGRADE = 66
FULL_HEIGHT_DEGRADE = 10


class SizeManager:
    """Class handling  the size reduction of the poezio UI."""

    def __init__(self, core: 'Core') -> None:
        self._core = core

    @property
    def tab_degrade_x(self) -> bool:
        """Returns True if we need to reduce the horizontal size used by the tab."""
        if base_wins.TAB_WIN is None:
            raise ValueError()
        _, x = base_wins.TAB_WIN.getmaxyx()
        return x < THRESHOLD_WIDTH_DEGRADE

    @property
    def tab_degrade_y(self) -> bool:
        """Returns True if we need to reduce the vertical size used by the tab."""
        if base_wins.TAB_WIN is None:
            raise ValueError()
        y, x = base_wins.TAB_WIN.getmaxyx()
        return y < THRESHOLD_HEIGHT_DEGRADE

    @property
    def core_degrade_x(self) -> bool:
        """Returns True if we need to reduce the horizontal size used by poezio."""
        if self._core.stdscr is None:
            return False
        _, x = self._core.stdscr.getmaxyx()
        return x < FULL_WIDTH_DEGRADE

    @property
    def core_degrade_y(self) -> bool:
        """Returns True if we need to reduce the vertical size used by the poezio."""
        if self._core.stdscr is None:
            return False
        y, _ = self._core.stdscr.getmaxyx()
        return y < FULL_HEIGHT_DEGRADE



"""
This plugin lets you execute a command, to notify you from new important
messages.

Installation and configuration
------------------------------

You need to create a plugin configuration file. Create a file named :file:`simple_notify.cfg`
into your plugins configuration directory (:file:`~/.config/poezio/plugins` by
default), and fill it like this:

First example:

.. code-block:: ini

    [simple_notify]
    command = notify-send -i /path/to/poezio/data/poezio_80.png "New message from %(from)s" "%(body)s"

Second example:

.. code-block:: ini

    [simple_notify]
    command = echo \\<%(from)s\\> %(body)s >> some.fifo
    delay = 3
    after_command = echo >> some.fifo

You can put any command, instead of these ones. You can also use the
special keywords ``%(from)s`` and ``%(body)s`` that will be replaced
directly in the command line by the author of the message, and the body.

The first example shown above will display something like this:

.. figure:: ../images/simple_notify_example.png
    :alt: Simple notify example

The second example will first write the author and the message in a
fifo, that fifo can locally be read by some other program (was tested
with the xmobar PipeReader command, which displays what is read from a
fifo into a status bar. Be careful, you have two different fifos in
that case, don’t get confused). The :term:`delay` and :term:`after_command` options
are used to erase/delete/kill the notification after a certain
delay.  In our example it is used to display an empty message in our
xmobar, erasing the notification after 3 seconds.

Third example:

.. code-block:: ini

    [simple_notify]
    command = notify-send -i /path/to/poezio/data/poezio_80.png "New message from %(from)s" "%(body)s"
    muc_too = true
    muc_list = someroom@conference.jabber.org:someotherroom@conference.jabber.org

If present and set to ``True``, the ``muc_too`` option will also trigger a
notification when a new message arrives on a Multi User Chat you've joined.

If present and set to a colon separated list of muc JIDs, muc_list together
with muc_too = true will only notify when a new message arrives on a Multi
User Chat, you've joined if it is present on the list.

.. note:: If you set the :term:`exec_remote` option to ``true`` into the
    main configuration file, the command will be executed remotely
    (as explained in the :ref:`link-plugin` plugin help).

Options defined
---------------

.. glossary::
    :sorted:

    command
        The command to execute (with special keywords ``%{from}s`` and ``${body}s``)

    delay
        Delay after which :term:`after_command` must be executed.

    after_command
        Command to run after :term:`delay`. You probably want to clean up things.

    muc_too
        Boolean indicating whether new messages in Multi User Chat rooms should
        trigger a notification or not.

"""

from poezio.plugin import BasePlugin
from poezio.xhtml import get_body_from_message_stanza
from poezio.timed_events import DelayedEvent
import shlex
from poezio import common, tabs
from slixmpp import Message


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('private_msg', self.on_private_msg)
        self.api.add_event_handler('conversation_msg',
                                   self.on_conversation_msg)
        if self.config.get('muc_too', False):
            self.api.add_event_handler('muc_msg', self.on_muc_msg)
        self.api.add_event_handler('highlight', self.on_highlight)

    def on_private_msg(self, message: Message, tab: tabs.PrivateTab) -> None:
        fro = message['from']
        self.do_notify(message, fro)

    def on_highlight(self, message: Message, tab: tabs.MucTab) -> None:
        whitelist = self.config.get('muc_list', '').split(':')
        # prevents double notifications
        if message['from'].bare in whitelist:
            return
        fro = message['from'].resource
        self.do_notify(message, fro)

    def on_conversation_msg(self, message: Message, tab: tabs.OneToOneTab) -> None:
        fro = message['from']
        if fro.bare != self.core.xmpp.boundjid.bare:
            self.do_notify(message, fro)

    def on_muc_msg(self, message: Message, tab: tabs.MucTab) -> None:
        # Don't notify if message is from yourself
        if message['from'].resource == tab.own_nick:
            return

        fro = message['from'].full
        muc = message['from'].bare
        whitelist = self.config.get('muc_list', '').split(':')

        # Prevent old messages to be notified
        # find_delayed_tag(message) returns (True, the datetime) or
        # (False, None)
        if not common.find_delayed_tag(message)[0]:
            # Only notify if whitelist is empty or muc in whitelist
            if whitelist == [''] or muc in whitelist:
                self.do_notify(message, fro)

    def do_notify(self, message: Message, fro: str) -> None:
        body = get_body_from_message_stanza(message, use_xhtml=False)
        if not body:
            return
        command_str = self.config.get('command', '').strip()
        if not command_str:
            self.api.information(
                'No notification command was provided in the configuration file',
                'Warning')
            return
        command = [
            arg % {
                'body': body.replace('\n', ' '),
                'from': fro
            } for arg in shlex.split(command_str)
        ]
        self.core.exec_command(command)
        after_command_str = self.config.get('after_command', '').strip()
        if not after_command_str:
            return
        after_command = [
            arg % {
                'body': body.replace('\n', ' '),
                'from': fro
            } for arg in shlex.split(after_command_str)
        ]
        delayed_event = DelayedEvent(
            self.config.get('delay', 1), self.core.exec_command, after_command)
        self.api.add_timed_event(delayed_event)



"""
Standalone functions used by the modules
"""

import string
from typing import Optional, List
from poezio.ui.consts import FORMAT_CHAR, FORMAT_CHARS

DIGITS = string.digits + '-'


def find_first_format_char(text: str,
                           chars: Optional[str] = None) -> int:
    """Find the first \x19 from the string."""
    to_find: str = chars or FORMAT_CHARS
    pos = -1
    for char in to_find:
        p = text.find(char)
        if p == -1:
            continue
        if pos == -1 or p < pos:
            pos = p
    return pos


def truncate_nick(nick: Optional[str], size: int = 10) -> str:
    """Cut up a nick up to ``size``."""
    if size < 1:
        size = 1
    if nick:
        if len(nick) > size:
            return nick[:size] + '…'
        return nick
    return ''


def parse_attrs(text: str, previous: Optional[List[str]] = None) -> List[str]:
    """Parse the various special formatting attributes in  a string."""
    next_attr_char = text.find(FORMAT_CHAR)
    if previous:
        attrs = previous
    else:
        attrs = []
    while next_attr_char != -1 and text:
        if next_attr_char + 1 < len(text):
            attr_char = text[next_attr_char + 1].lower()
        else:
            attr_char = '\0'
        if attr_char == 'o':
            attrs = []
        elif attr_char == 'u':
            attrs.append('u')
        elif attr_char == 'b':
            attrs.append('b')
        elif attr_char == 'i':
            attrs.append('i')
        if attr_char in DIGITS and attr_char:
            color_str = text[next_attr_char + 1:text.find('}', next_attr_char)]
            if color_str:
                attrs.append(color_str + '}')
            text = text[next_attr_char + len(color_str) + 2:]
        else:
            text = text[next_attr_char + 2:]
        next_attr_char = text.find(FORMAT_CHAR)
    return attrs



"""
Upload a file and auto-complete the input with its URL.

Usage
-----

This plugin adds a command to the chat tabs.

.. glossary::

    /upload
        **Usage:** ``/upload <filename>``

        Uploads the <filename> file to the preferred HTTP File Upload
        service (see XEP-0363) and fill the input with its URL.


"""

from typing import cast

import traceback
from os.path import expanduser
from glob import glob
from pathlib import Path

from slixmpp.plugins.xep_0363.http_upload import FileTooBig, HTTPError, UploadServiceNotFound

from poezio.plugin import BasePlugin
from poezio.core.structs import Completion
from poezio.decorators import command_args_parser
from poezio.windows.inputs import Input
from poezio import tabs


class HTTPUploadNotFound(Exception):
    def __init__(self) -> None:
        super().__init__('slixmpp XEP-0363 plugin failed to load')


class Plugin(BasePlugin):
    dependencies = {'embed'}

    def init(self) -> None:
        self.embed = self.refs['embed']

        if not self.core.xmpp['xep_0363']:
            raise HTTPUploadNotFound()
        if not self.core.xmpp['xep_0453']:
            self.api.information(
                'slixmpp XEP-0454 plugin failed to load. '
                'Will not be able to encrypt uploaded files.',
                'Warning',
            )
        for _class in (tabs.PrivateTab, tabs.StaticConversationTab, tabs.DynamicConversationTab, tabs.MucTab):
            self.api.add_tab_command(
                _class,
                'upload',
                self.command_upload,
                usage='<filename>',
                help='Upload a file and auto-complete the input with its URL.',
                short='Upload a file',
                completion=self.completion_filename)

    async def upload(self, filename: Path, encrypted: bool = False) -> str | None:
        try:
            upload_file = self.core.xmpp['xep_0363'].upload_file
            if encrypted:
                upload_file = self.core.xmpp['xep_0454'].upload_file
            url = cast(str, await upload_file(filename))
        except UploadServiceNotFound:
            self.api.information('HTTP Upload service not found.', 'Error')
            return None
        except (FileTooBig, HTTPError) as exn:
            self.api.information(str(exn), 'Error')
            return None
        except Exception:
            exception = traceback.format_exc()
            self.api.information('Failed to upload file: %s' % exception,
                                 'Error')
            return None
        return url

    async def send_upload(self, filename: Path, tab: tabs.ChatTab,
                          encrypted: bool = False) -> None:
        url = await self.upload(filename, encrypted)
        if url is not None:
            self.embed.embed_image_url(url, tab)

    @command_args_parser.quoted(1)
    def command_upload(self, args: list[str]) -> None:
        if not args:
            self.core.command.help('upload')
            return
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        filename_str, = args
        filename = Path(expanduser(filename_str))
        encrypted = self.core.xmpp['xep_0454'] and tab.e2e_encryption is not None
        self.core.register_task(self.send_upload(filename, tab, encrypted))

    @staticmethod
    def completion_filename(the_input: Input) -> Completion:
        txt = expanduser(the_input.get_text()[8:])
        files = glob(txt + '*')
        return Completion(the_input.auto_completion, files, quotify=False)



"""
Lists old versions of a corrected message, or retracted/moderated messages.

Moderated and retracted messages only have one version, obviously.

Usage
-----

.. glossary::

    /display_corrections
        **Usage:** ``/display_corrections [number]``

        This command lists the old versions of a message.

        Without argument, it will list the last corrected message if there
        is any. If you give an integer as an argument, ``/display_corrections``
        will go back gradually in the buffer to find the message matching
        that number (starting from 1, for the last corrected message).

        If you are scrolling in the buffer, Poezio will list the corrected messages
        starting from the first you can see.  (although there are some problems with
        multiline messages).


"""
from poezio.plugin import BasePlugin
from poezio.common import shell_split
from poezio import tabs
from poezio.ui.types import Message, DeletedMessage
from poezio.theming import get_theme


class Plugin(BasePlugin):
    def init(self) -> None:
        for tab_type in (tabs.MucTab, tabs.PrivateTab, tabs.DynamicConversationTab, tabs.StaticConversationTab):
            self.api.add_tab_command(
                tab_type,
                'display_corrections',
                handler=self.command_display_corrections,
                usage='<number>',
                help=(
                    'Display all the corrections of the number-th last '
                    'corrected message.'
                ),
                short='Display the corrections of a message')

    def find_corrected(self, nb: int) -> Message | None:
        messages = self.api.get_conversation_messages()
        if not messages:
            return None
        for message in reversed(messages):
            if not isinstance(message, (Message, DeletedMessage)):
                continue
            if message.old_message:
                if nb == 1:
                    # Corrected
                    if isinstance(message, Message):
                        return message
                    # Retracted
                    else:
                        return message.old_message
                else:
                    nb -= 1
        return None

    def command_display_corrections(self, args: str) -> None:
        theme = get_theme()
        list_args = shell_split(args)
        if len(list_args) == 1:
            try:
                nb = int(args[0])
            except ValueError:
                self.api.run_command('/help display_corrections')
                return
        else:
            nb = 1
        message = self.find_corrected(nb)
        if message:
            display = []
            while message:
                str_time = message.time.strftime(theme.SHORT_TIME_FORMAT)
                display.append('%s %s%s%s %s' %
                               (str_time, '* '
                                if message.me else '', message.nickname, ''
                                if message.me else '>', message.txt))
                message = message.old_message
            self.api.information(
                'Older versions:\n' + '\n'.join(display[::-1]), 'Info')
        else:
            self.api.information('No corrected message found.', 'Warning')

    def cleanup(self) -> None:
        del self.config



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Various methods to convert
shell colors to poezio colors,
xhtml code to shell colors,
poezio colors to xhtml code
"""

import hashlib
import re
from base64 import b64encode, b64decode
from os import path
from urllib.parse import unquote
from pathlib import Path

from io import BytesIO
from xml import sax
from xml.sax import saxutils
from xml.sax.handler import ContentHandler
from xml.sax.xmlreader import AttributesNSImpl
from typing import List, Dict, Optional, Union, Tuple, cast

from slixmpp import Message
from slixmpp.xmlstream import ET
from poezio.config import config
from poezio.colors import ncurses_color_to_rgb

digits = '0123456789'  # never trust the modules

XHTML_NS = 'http://www.w3.org/1999/xhtml'

# HTML named colors
colors: dict[str, int] = {
    'aliceblue': 231,
    'antiquewhite': 231,
    'aqua': 51,
    'aquamarine': 122,
    'azure': 231,
    'beige': 231,
    'bisque': 230,
    'black': 232,
    'blanchedalmond': 230,
    'blue': 21,
    'blueviolet': 135,
    'brown': 124,
    'burlywood': 223,
    'cadetblue': 109,
    'chartreuse': 118,
    'chocolate': 172,
    'coral': 209,
    'cornflowerblue': 111,
    'cornsilk': 231,
    'crimson': 197,
    'cyan': 51,
    'darkblue': 19,
    'darkcyan': 37,
    'darkgoldenrod': 178,
    'darkgray': 247,
    'darkgreen': 28,
    'darkgrey': 247,
    'darkkhaki': 186,
    'darkmagenta': 127,
    'darkolivegreen': 65,
    'darkorange': 214,
    'darkorchid': 134,
    'darkred': 124,
    'darksalmon': 216,
    'darkseagreen': 151,
    'darkslateblue': 61,
    'darkslategray': 59,
    'darkslategrey': 59,
    'darkturquoise': 44,
    'darkviolet': 128,
    'deeppink': 199,
    'deepskyblue': 45,
    'dimgray': 241,
    'dimgrey': 241,
    'dodgerblue': 39,
    'firebrick': 160,
    'floralwhite': 231,
    'forestgreen': 34,
    'fuchsia': 201,
    'gainsboro': 252,
    'ghostwhite': 231,
    'gold': 226,
    'goldenrod': 214,
    'gray': 244,
    'green': 34,
    'greenyellow': 191,
    'grey': 244,
    'honeydew': 231,
    'hotpink': 212,
    'indianred': 174,
    'indigo': 55,
    'ivory': 231,
    'khaki': 229,
    'lavender': 231,
    'lavenderblush': 231,
    'lawngreen': 118,
    'lemonchiffon': 230,
    'lightblue': 195,
    'lightcoral': 217,
    'lightcyan': 231,
    'lightgoldenrodyellow': 230,
    'lightgray': 251,
    'lightgreen': 157,
    'lightgrey': 251,
    'lightpink': 224,
    'lightsalmon': 216,
    'lightseagreen': 43,
    'lightskyblue': 153,
    'lightslategray': 109,
    'lightslategrey': 109,
    'lightsteelblue': 189,
    'lightyellow': 231,
    'lime': 46,
    'limegreen': 77,
    'linen': 231,
    'magenta': 201,
    'maroon': 124,
    'mediumaquamarine': 115,
    'mediumblue': 20,
    'mediumorchid': 170,
    'mediumpurple': 141,
    'mediumseagreen': 78,
    'mediumslateblue': 105,
    'mediumspringgreen': 49,
    'mediumturquoise': 80,
    'mediumvioletred': 163,
    'midnightblue': 18,
    'mintcream': 231,
    'mistyrose': 231,
    'moccasin': 230,
    'navajowhite': 230,
    'navy': 19,
    'oldlace': 231,
    'olive': 142,
    'olivedrab': 106,
    'orange': 214,
    'orangered': 202,
    'orchid': 213,
    'palegoldenrod': 229,
    'palegreen': 157,
    'paleturquoise': 195,
    'palevioletred': 211,
    'papayawhip': 231,
    'peachpuff': 230,
    'peru': 179,
    'pink': 224,
    'plum': 219,
    'powderblue': 195,
    'purple': 127,
    'red': 196,
    'rosybrown': 181,
    'royalblue': 69,
    'saddlebrown': 130,
    'salmon': 216,
    'sandybrown': 216,
    'seagreen': 72,
    'seashell': 231,
    'sienna': 131,
    'silver': 250,
    'skyblue': 153,
    'slateblue': 104,
    'slategray': 109,
    'slategrey': 109,
    'snow': 231,
    'springgreen': 48,
    'steelblue': 74,
    'tan': 187,
    'teal': 37,
    'thistle': 225,
    'tomato': 209,
    'turquoise': 86,
    'violet': 219,
    'wheat': 230,
    'white': 255,
    'whitesmoke': 255,
    'yellow': 226,
    'yellowgreen': 149
}

whitespace_re = re.compile(r'\s+')

xhtml_attr_re = re.compile(r'\x19-?\d[^}]*}|\x19[buaio]')
xhtml_data_re = re.compile(r'data:image/([a-z]+);base64,(.+)')
xhtml_cid_re = re.compile(r'^cid:(sha1\+[0-9a-fA-F]+@bob\.xmpp\.org)$')
poezio_color_double = re.compile(r'(?:\x19\d+}|\x19\d)+(\x19\d|\x19\d+})')
poezio_format_trim = re.compile(r'(\x19\d+}|\x19\d|\x19[buaio]|\x19o)+\x19o')

xhtml_simple_attr_re = re.compile(r'\x19\d')


def get_body_from_message_stanza(message: Message,
                                 use_xhtml: bool = False,
                                 extract_images_to: Path | None = None
                                 ) -> str:
    """
    Returns a string with xhtml markups converted to
    poezio colors if there's an xhtml_im element, or
    the body (without any color) otherwise
    """
    if not use_xhtml:
        return cast(str, message['body'])
    xhtml = message.xml.find('{http://jabber.org/protocol/xhtml-im}html')
    if xhtml is None:
        return cast(str, message['body'])
    xhtml_body = xhtml.find('{http://www.w3.org/1999/xhtml}body')
    if xhtml_body is None:
        return cast(str, message['body'])
    content = xhtml_to_poezio_colors(xhtml_body, tmp_dir=extract_images_to)
    content = content if content else message['body']
    return content or " "


def rgb_to_html(rgb: Tuple[float, float, float]) -> str:
    """Get the RGB HTML value"""
    r, g, b = rgb
    return '#%02X%02X%02X' % (round(r * 255), round(g * 255), round(b * 255))


def ncurses_color_to_html(color: int) -> str:
    """
    Takes an int between 0 and 256 and returns
    a string of the form #XXXXXX representing an
    html color.
    """
    return rgb_to_html(ncurses_color_to_rgb(color))


def _parse_css_color(name: str) -> int:
    """Parse a css color, either in name or in various value types."""
    if name[0] == '#':
        name = name[1:]
        length = len(name)
        if length != 3 and length != 6:
            return -1
        value = int(name, 16)
        if length == 6:
            r = value >> 16
            g = (value >> 8) & 0xff
            b = value & 0xff
            if r == g == b:
                return int(232 + 0.0941 * r)
            mult = 0.0235
        else:
            r = value >> 8
            g = (value >> 4) & 0xf
            b = value & 0xf
            if r == g == b:
                return int(232 + 1.54 * r)
            mult = 0.3984
        return 6 * 6 * int(mult * r) + 6 * int(mult * g) + int(mult * b) + 16
    if name in colors:
        return colors[name]
    return -1


def _parse_css(css: str) -> str:
    """Parse a reduced subset of CSS attributes."""
    shell = ''
    rules = css.split(';')
    for rule in rules:
        if ':' not in rule:
            continue
        key, value = rule.split(':', 1)
        key = key.strip()
        value = value.strip()
        if key == 'background-color':
            pass  # shell += '\x191'
        elif key == 'color':
            color = _parse_css_color(value)
            if color != -1:
                shell += '\x19%d}' % color
        elif key == 'font-style':
            shell += '\x19i'
        elif key == 'font-weight':
            shell += '\x19b'
        elif key == 'margin-left':
            shell += '    '
        elif key == 'text-align':
            pass
        elif key == 'text-decoration':
            if value == 'underline':
                shell += '\x19u'
            elif value == 'blink':
                shell += '\x19a'
    return shell


def _trim(string: str) -> str:
    """Remove  whitespace."""
    return re.sub(whitespace_re, ' ', string)


def get_hash(data: bytes) -> str:
    """Get the hash for the given data.

    Currently using SHA-256, this might change in the future.
    base64 gives shorter hashes than hex, so use that.
    """
    return b64encode(hashlib.sha256(data).digest()).rstrip(b'=').replace(
        b'/', b'-').decode()


class XHTMLHandler(ContentHandler):
    def __init__(self, force_ns: bool = False,
                 tmp_image_dir: Path | None = None) -> None:
        self.builder: List[str] = []
        self.formatting: List[str] = []
        self.attrs: List[Dict[str, str]] = []
        self.list_state: List[Union[str, int]] = []
        self.cids: Dict[str, Optional[str]] = {}
        self.is_pre = False
        self.a_start = 0
        # do not care about xhtml-in namespace
        self.force_ns = force_ns

        self.tmp_image_dir = Path(tmp_image_dir) if tmp_image_dir else None
        self.enable_css_parsing = config.getbool('enable_css_parsing')

    @property
    def result(self) -> str:
        sanitized = re.sub(poezio_color_double, r'\1',
                           ''.join(self.builder).strip())
        return re.sub(poezio_format_trim, '\x19o', sanitized)

    def append_formatting(self, formatting: str) -> None:
        self.formatting.append(formatting)
        self.builder.append(formatting)

    def pop_formatting(self) -> None:
        self.formatting.pop()
        self.builder.append('\x19o' + ''.join(self.formatting))

    def characters(self, characters: str) -> None:
        self.builder.append(characters if self.is_pre else _trim(characters))

    def startElementNS(self, name: tuple[str | None, str],
                       qname: str | None, attrs: AttributesNSImpl) -> None:
        if name[0] != XHTML_NS and not self.force_ns:
            return

        builder = self.builder
        cleaned_attrs = {
            name: value
            for ((ns, name), value) in attrs.items() if ns is None
        }
        self.attrs.append(cleaned_attrs)

        if 'style' in cleaned_attrs and self.enable_css_parsing:
            style = _parse_css(cleaned_attrs['style'])
            self.append_formatting(style)

        attr_name = name[1]
        if attr_name == 'a':
            self.append_formatting('\x19u')
            self.a_start = len(self.builder)
        elif attr_name == 'blockquote':
            builder.append('“')
        elif attr_name == 'br':
            builder.append('\n')
        elif attr_name == 'cite':
            self.append_formatting('\x19u')
        elif attr_name == 'em':
            self.append_formatting('\x19i')
        elif attr_name == 'img':
            if re.match(xhtml_data_re,
                        cleaned_attrs['src']) and self.tmp_image_dir is not None:
                type_, data = [
                    i for i in re.split(xhtml_data_re, cleaned_attrs['src']) if i
                ]
                bin_data = b64decode(unquote(data))
                filename = get_hash(bin_data) + '.' + type_
                filepath = self.tmp_image_dir / filename
                if not path.exists(filepath):
                    try:
                        self.tmp_image_dir.mkdir(parents=True, exist_ok=True)
                        with open(filepath, 'wb') as fd:
                            fd.write(bin_data)
                        builder.append('[file stored as %s]' % filename)
                    except Exception as e:
                        builder.append('[Error while saving image: %s]' % e)
                else:
                    builder.append('[file stored as %s]' % filename)
            elif re.match(xhtml_cid_re,
                          cleaned_attrs['src']) and self.tmp_image_dir is not None:
                cid = cleaned_attrs['src'][4:]
                # TODO: start the download outside of this parser.
                self.cids[cid] = None
                builder.append('[Error: XEP-0231 file transfer not yet supported]')
            else:
                builder.append(_trim(cleaned_attrs['src']))
            if 'alt' in cleaned_attrs:
                builder.append(' (%s)' % _trim(cleaned_attrs['alt']))
        elif attr_name == 'ul':
            self.list_state.append('ul')
        elif attr_name == 'ol':
            self.list_state.append(1)
        elif attr_name == 'li':
            try:
                state = self.list_state[-1]
            except IndexError:
                state = 'ul'
            if state == 'ul':
                builder.append('\n• ')
            elif isinstance(state, int):
                builder.append('\n%s) ' % state)
                state += 1
                self.list_state[-1] = state
        elif attr_name == 'p':
            builder.append('\n')
        elif attr_name == 'pre':
            builder.append('\n')
            self.is_pre = True
        elif attr_name == 'strong':
            self.append_formatting('\x19b')

    def endElementNS(self, name: tuple[str | None, str], qname: str | None) -> None:
        if name[0] != XHTML_NS and not self.force_ns:
            return

        builder = self.builder
        attrs = self.attrs.pop()
        attr_name = name[1]

        if attr_name == 'a':
            self.pop_formatting()
            # do not display the link twice
            text_elements = [
                x for x in self.builder[self.a_start:]
                if not x.startswith('\x19')
            ]
            link_text = ''.join(text_elements).strip()
            if 'href' in attrs and attrs['href'] != link_text:
                builder.append(' (%s)' % _trim(attrs['href']))
        elif attr_name == 'blockquote':
            builder.append('”\n')
        elif attr_name in ('cite', 'em', 'strong'):
            self.pop_formatting()
        elif attr_name in ('ol', 'p', 'ul'):
            builder.append('\n')
        elif attr_name == 'pre':
            builder.append('\n')
            self.is_pre = False

        if 'style' in attrs and self.enable_css_parsing:
            self.pop_formatting()

        if 'title' in attrs:
            builder.append(' [' + attrs['title'] + ']')


def xhtml_to_poezio_colors(xml: str | bytes | ET.Element, force: bool = False,
                           tmp_dir: Path | None = None) -> str:
    if isinstance(xml, str):
        xml = xml.encode('utf8')
    elif not isinstance(xml, bytes):
        xml = ET.tostring(xml)

    handler = XHTMLHandler(force_ns=force, tmp_image_dir=tmp_dir)
    parser = sax.make_parser()
    parser.setFeature(sax.handler.feature_namespaces, True)
    parser.setContentHandler(handler)
    parser.parse(BytesIO(xml))
    return handler.result


def clean_text(s: str) -> str:
    """
    Remove all xhtml-im attributes (\x19etc) from the string with the
    complete color format, i.e \x19xxx}
    """
    s = re.sub(xhtml_attr_re, "", s)
    return s


def clean_text_simple(string: str) -> str:
    """
    Remove all \x19 from the string formatted with simple colors:
    \x198
    """
    pos = string.find('\x19')
    while pos != -1:
        string = string[:pos] + string[pos + 2:]
        pos = string.find('\x19')
    return string


def convert_simple_to_full_colors(text: str) -> str:
    """
    takes a \x19n formatted string and returns
    a \x19n} formatted one.
    """
    # TODO, have a single list of this. This is some sort of
    # duplicate from ui.consts.FORMAT_CHARS
    mapping = str.maketrans({
        '\x0E': '\x19b',
        '\x0F': '\x19o',
        '\x10': '\x19u',
        '\x11': '\x191',
        '\x12': '\x192',
        '\x13': '\x193',
        '\x14': '\x194',
        '\x15': '\x195',
        '\x16': '\x196',
        '\x17': '\x197',
        '\x18': '\x198',
        '\x19': '\x199',
        '\x1A': '\x19i'
    })
    text = text.translate(mapping)

    def add_curly_bracket(match: re.Match[str]) -> str:
        return match.group(0) + '}'

    return re.sub(xhtml_simple_attr_re, add_curly_bracket, text)


number_to_color_names: Dict[int, str] = {
    1: 'red',
    2: 'green',
    3: 'yellow',
    4: 'blue',
    5: 'violet',
    6: 'turquoise',
    7: 'white'
}


def format_inline_css(_dict: Dict[str, str]) -> str:
    return ''.join(('%s: %s;' % (key, value) for key, value in _dict.items()))


def poezio_colors_to_html(string: str) -> str:
    """
    Convert poezio colors to html
    (e.g. \x191}: <span style='color: red'>)
    """
    # Maintain a list of the current css attributes used
    # And check if a tag is open (by design, we only open
    # spans tag, and they cannot be nested.
    current_attrs: Dict[str, str] = {}
    tag_open = False
    next_attr_char = string.find('\x19')
    build = ["<body xmlns='http://www.w3.org/1999/xhtml'><p>"]

    def check_property(key: str, value: str) -> None:
        nonlocal tag_open
        if current_attrs.get(key, None) == value:
            return
        current_attrs[key] = value
        if tag_open:
            tag_open = False
            build.append('</span>')

    while next_attr_char != -1:
        attr_char = string[next_attr_char + 1].lower()

        if next_attr_char != 0 and string[:next_attr_char]:
            if current_attrs and not tag_open:
                build.append(
                    '<span style="%s">' % format_inline_css(current_attrs))
                tag_open = True
            build.append(saxutils.escape(string[:next_attr_char]))

        if attr_char == 'o':
            if tag_open:
                build.append('</span>')
                tag_open = False
            current_attrs = {}
        elif attr_char == 'b':
            check_property('font-weight', 'bold')
        elif attr_char == 'u':
            check_property('text-decoration', 'underline')
        elif attr_char == 'i':
            check_property('font-style', 'italic')

        if attr_char in digits:
            number_str = string[next_attr_char +
                                1:string.find('}', next_attr_char)]
            number = int(number_str)
            if number in number_to_color_names:
                check_property('color',
                               number_to_color_names.get(number, 'black'))
            else:
                check_property('color', ncurses_color_to_html(number))
            string = string[next_attr_char + len(number_str) + 2:]
        else:
            string = string[next_attr_char + 2:]
        next_attr_char = string.find('\x19')

    if current_attrs and not tag_open and string:
        build.append('<span style="%s">' % format_inline_css(current_attrs))
        tag_open = True
    build.append(saxutils.escape(string))
    if tag_open:
        build.append('</span>')
    build.append("</p></body>")
    text = ''.join(build)
    return text.replace('\n', '<br />')



"""
Replace some word with some other word in a message before sending it.

Configuration example
---------------------
.. code-block:: ini

[replace_word]
# How to appear casual in your daily conversations.
yes = yep
no = nope

Usage
-----
Just use the word in a message. It will be replaced automatically.

"""

from poezio.plugin import BasePlugin
from slixmpp import Message
from typing import Any
import re


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('conversation_say', self.replace_pattern)
        self.api.add_event_handler('muc_say', self.replace_pattern)
        self.api.add_event_handler('private_say', self.replace_pattern)

    def replace_pattern(self, message: Message, tab: Any) -> None:
        """
        Look for a given word in the message and replace it by the corresponding word.
        """
        body = message['body']
        for before in self.config.options("replace_word"):
            after = self.config.get(before, before)
            body = re.sub(r"\b%s\b" % before, after, body)
        message['body'] = body



"""
This plugin adds a ``/server_part`` command to leave all rooms
on a server.

Command
-------

.. glossary::

    /server_part
        **Usage:** ``/server_part [<server> [message]]``

        Leave all rooms on ``<server>``, if not provided and the current
        tab is a chatroom tab, it will leave all rooms on the current server.
        ``[message]`` can indicate a quit message.


"""
from slixmpp import JID, InvalidJID
from poezio.plugin import BasePlugin
from poezio.tabs import MucTab
from poezio.decorators import command_args_parser
from poezio.core.structs import Completion
from poezio.windows.inputs import Input


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'server_part',
            self.command_server_part,
            usage='[<server> [message]]',
            short='Leave all the rooms on a server',
            help='Leave all the rooms on a sever.',
            completion=self.completion_server_part)

    @command_args_parser.quoted(0, 2, defaults=[])
    def command_server_part(self, args: list[str]) -> None:
        current_tab = self.api.current_tab()
        message = ''
        if not args:
            if not isinstance(current_tab, MucTab):
                self.core.command.help('server_part')
                return
            jid = current_tab.jid.bare
        elif len(args) == 1:
            try:
                jid = JID(args[0]).domain
            except InvalidJID:
                self.core.command.help('server_part')
                return
        else:
            try:
                jid = JID(args[0]).domain
            except InvalidJID:
                self.core.command.help('server_part')
                return
            message = args[1]

        for tab in self.core.get_tabs(MucTab):
            if tab.name.endswith(jid):
                tab.command_part(message)

    def completion_server_part(self, the_input: Input) -> Completion:
        serv_list = set()
        for tab in self.core.get_tabs(MucTab):
            if tab.joined:
                serv = tab.jid.server
                serv_list.add(serv)
        return Completion(the_input.new_completion, sorted(serv_list), 1, ' ')



"""

This plugins allows commands to be sent to poezio via a named pipe.

You can run the same commands that you would in the poezio input
(e.g. ``echo '/message toto@example.tld Hi' >> /tmp/poezio.fifo``).

Configuration
-------------

.. glossary::
    :sorted:

    pipename
        **Default:** :file:`/tmp/poezio.fifo`

        The path to the fifo which will receive commands.

"""

from poezio.plugin import BasePlugin
import os
import stat
import logging
import asyncio

log = logging.getLogger(__name__)

PIPENAME = "/tmp/poezio.fifo"


class Plugin(BasePlugin):
    def init(self) -> None:
        self.stop = False

        self.pipename = self.config.get("pipename", PIPENAME)

        if not os.path.exists(self.pipename):
            os.mkfifo(self.pipename)

        if not stat.S_ISFIFO(os.stat(self.pipename).st_mode):
            raise TypeError("File %s is not a fifo file" % self.pipename)

        self.fd = os.open(self.pipename, os.O_RDONLY | os.O_NONBLOCK)

        self.data = b""
        asyncio.get_event_loop().add_reader(self.fd, self.read_from_fifo)

    def read_from_fifo(self) -> None:
        data = os.read(self.fd, 512)
        if not data:
            # EOF, close the fifo. And reopen it
            asyncio.get_event_loop().remove_reader(self.fd)
            os.close(self.fd)
            self.fd = os.open(self.pipename, os.O_RDONLY | os.O_NONBLOCK)
            asyncio.get_event_loop().add_reader(self.fd, self.read_from_fifo)
            self.data = b''
        else:
            self.data += data
        split_data = self.data.split(b'\n', 1)
        if len(split_data) == 2:
            line, self.data = split_data
            log.debug("run: %s" % (line.decode().strip()))
            self.api.run_command(line.decode().strip())

    def cleanup(self) -> None:
        asyncio.get_event_loop().remove_reader(self.fd)
        os.close(self.fd)
